import numpy as np
import scipy.sparse as sp
# import scipy.sparse.linalg as splinalg
import matplotlib.pyplot as plt
from PIL import Image, ImageFilter
from scipy.ndimage import map_coordinates


def im2arr(img):
    """pillow grayscale image to normalized numpy array"""
    return np.array(img, dtype=np.float64) / 255.0


def arr2im(arr):
    """normalized numpy array pillow grayscale image"""
    return Image.fromarray(arr * 255).convert('L')


def imshowpair(img0, img1):
    """Show blended grayscale images. Channel map: img0 -> g; img1 -> (r, b)"""
    imgblend = np.stack([img1, img0, img1], axis=2)
    Image.fromarray((imgblend * 255).astype(np.uint8))
    plt.imshow(imgblend)


# reference: https://github.com/python/cpython/blob/master/Lib/colorsys.py
def hs1_to_rgb(h, s):
    """numpy array version of HSV to RGB converter with V=1"""
    r, g, b = np.ones_like(h), np.ones_like(h), np.ones_like(h)
    # compute some "magical" values
    i = (h * 6.0).astype(np.int8)
    f = (h * 6.0) - i
    p = 1.0 - s
    q = 1.0 - s * f
    t = 1.0 - s * (1.0 - f)
    i = i % 6
    # i = 0: v=1, t, p
    g[i == 0], b[i == 0] = t[i == 0], p[i == 0]
    # i = 1: q, v=1, p
    r[i == 1], b[i == 1] = q[i == 1], p[i == 1]
    # i = 2: p, v=1, t
    r[i == 2], b[i == 2] = p[i == 2], t[i == 2]
    # i = 3: p, q, v=1
    r[i == 3], g[i == 3] = p[i == 3], q[i == 3]
    # i = 4: t, p, v=1
    r[i == 4], g[i == 4] = t[i == 4], p[i == 4]
    # i = 5: v=1, p, q
    g[i == 5], b[i == 5] = p[i == 5], q[i == 5]

    return np.stack([r, g, b], axis=2)


def flowshow_quiver(ux, uy):
    """visualize the optical flow as quiver plot"""
    plt.axis('equal')
    plt.quiver(ux[::-1, :], uy[::-1, :])


def flowshow_color(ux, uy):
    """visualize the optical flow using color"""
    flow_norm = np.sqrt(ux ** 2 + uy ** 2)
    factor = np.max(flow_norm)
    angle = np.arctan2(uy, -ux)
    s = flow_norm / factor
    h = (angle + np.pi) / (2 * np.pi)
    colorimg = hs1_to_rgb(h, s)
    plt.imshow((colorimg*255).astype(np.uint8))


def show_flow_repr():
    """this shows an example of the color encoding of flow vectors"""
    uy, ux = np.mgrid[-1:1.01:0.1, -1:1.01:0.1]
    uy = -uy
    plt.figure('color representation of flow vector')
    plt.subplot(121)
    plt.axis('equal')
    flowshow_quiver(ux, uy)
    plt.subplot(122)
    flowshow_color(ux, uy)
    plt.draw()


def imresize(imgarr, size, smoothradius=None):
    """downscale image array"""
    if smoothradius is None:
        img = arr2im(imgarr)
    else:
        img = arr2im(imgarr).filter(
                ImageFilter.GaussianBlur(radius=smoothradius))
    return im2arr(img.resize(size, Image.BICUBIC))


def flowresize(ux, uy, newsize):
    """resize the optical flow"""
    wn, hn = newsize
    ho, wo = ux.shape
    ux = np.array(Image.fromarray(ux*wn/wo).resize(newsize, Image.BILINEAR))
    uy = np.array(Image.fromarray(uy*hn/ho).resize(newsize, Image.BILINEAR))
    return ux, uy


def imwarp(imgarr, ux, uy):
    """warp the images according to the optical flow"""
    h, w = imgarr.shape
    coord_invmap_src = np.mgrid[0.0:h, 0.0:w]
    coord_invmap_src[0, :, :] += uy
    coord_invmap_src[1, :, :] += ux
#    return transform.warp(imgarr, coord_invmap_src)
    return map_coordinates(imgarr, coord_invmap_src, order=1)


def mat2vec(m):
    """flatten matrix m"""
    return m.T.flatten()


def vec2mat(v, shape):
    """unflatten vectorized matrix v"""
    h, w = shape
    return v.reshape(w, h).T


# This decorator is added only for efficiency propose
class memoize_last(object):
    """decorator to cache the last value of a function."""
    def __init__(self, func):
        self.func = func
        self._stored_key = None
        self._stored_value = None

    def __call__(self, *args):
        if self._stored_key != args:
            self._stored_value = self.func(*args)
            self._stored_key = args
        return self._stored_value


@memoize_last
def Dx(h, w):
    """build sparse Dx operator"""
    e_1 = np.ones(w)
    e0 = -np.ones(w)
    e0[-1] = 0
    Dx_tilde = sp.spdiags([e_1, e0], [-1, 0], w, w)
    return sp.kron(Dx_tilde.T, sp.eye(h))


@memoize_last
def Dy(h, w):
    """build sparse Dy operator"""
    e1 = np.ones(h)
    e0 = -np.ones(h)
    e0[-1] = 0
    Dy_tilde = sp.spdiags([e0, e1], [0, 1], h, h)
    return sp.kron(sp.eye(w), Dy_tilde)


def D(imgvec, h, w):
    """compute image gradient along x,y directions. return shape (h*w, 2)"""
    return np.stack([Dx(h, w) @ imgvec,  Dy(h, w) @ imgvec], axis=1)


@memoize_last
def nabla(h, w):
    """build (sparse) nabla operator"""
    return sp.vstack([Dx(h, w),  Dy(h, w)])


def update_flow(i0, i1w, ux, uy, lamb, nbiter_pdhg=30):
    """main function that updates the optical flow for each step"""
    # TODO: implement this
    raise NotImplementedError


if __name__ == '__main__':
    img_dir = './'  # point to image folder
    # image examples are from http://vision.middlebury.edu/flow/data/
    imgname0, imgname1 = img_dir+'frame10.png', img_dir+'frame11.png'
    levels = (4, 2, 1)  # downscale factors for each scale
    nb_warps = 5  # number of linearized warping steps
    nbiter_pdhg = 100  # number of pdhg iterations to perform
    lamb = 7.0  # lambda coefficient
    visualize_stepwise_energies = False
    visualize_stepwise_updates = False

    # load and display normalized images
    img0 = im2arr(Image.open(imgname0).convert('L'))
    img1 = im2arr(Image.open(imgname1).convert('L'))
    height, width = img0.shape

#    # here we downsample for faster runtime
#    height, width = height // 2, width // 2
#    img0 = imresize(img0, (width, height), 2)
#    img1 = imresize(img1, (width, height), 2)
#    height, width = img0.shape
    plt.figure('before optical flow')
    imshowpair(img0, img1)
    plt.show()

    # initialize flow in x, y directions
    ux, uy = np.zeros((2, 2)), np.zeros((2, 2))

    # mutli-scale iterate coarse-to-fine
    for level in levels:
        print('downscale factor {}'.format(level))
        # resize images to current level
        h, w = height // level, width // level
        i0 = imresize(img0, (w, h), level)
        i1 = imresize(img1, (w, h), level)
        # upscale previous flow
        ux, uy = flowresize(ux, uy, (w, h))
        # perform multiple linearized warping
        for i in range(nb_warps):
            print('- warp step {}'.format(i))
            # get warped image
            i1w = imwarp(i1, ux, uy)
            # if required, draw status before update
            if visualize_stepwise_updates:
                plt.figure('before/after optim')
                plt.subplot(131)
                imshowpair(i0, i1w)
            # update flow using PDHG
            ux, uy, energies = update_flow(i0, i1w, ux, uy, lamb, nbiter_pdhg)
            # if required, draw status after update and the optical flow
            if visualize_stepwise_updates:
                plt.subplot(132)
                imshowpair(i0, imwarp(i1, ux, uy))
                plt.subplot(133)
                flowshow_color(ux, uy)
            # if required, plot evolution of primal energy
            if visualize_stepwise_energies:
                plt.figure('primal energy')
                plt.plot(range(len(energies)), energies)
            # show drawings when required
            if visualize_stepwise_updates or visualize_stepwise_energies:
                plt.show()

    plt.figure('before/after optical flow')
    plt.subplot(131)
    imshowpair(img0, img1)
    plt.subplot(132)
    img1w = imwarp(img1, ux, uy)
    imshowpair(img0, img1w)
    plt.subplot(133)
    flowshow_color(ux, uy)
    plt.show()
